﻿using EonaCat.VolumeMixer.Helpers;
using EonaCat.VolumeMixer.Interfaces;
using EonaCat.VolumeMixer.Models;
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Threading.Tasks;

namespace EonaCat.VolumeMixer.Managers
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.
    public class VolumeMixerManager : IDisposable
    {
        private const int VT_UI4 = 0x13;
        private readonly IMultiMediaDeviceEnumerator _deviceEnumerator;
        private bool _isDisposed = false;
        private readonly object _syncLock = new();

        private static readonly PropertyKey PKEY_Device_FriendlyName = new PropertyKey(new Guid("a45c254e-df1c-4efd-8020-67d146a850e0"), 14);
        private static readonly PropertyKey PKEY_AudioEndpoint_FormFactor = new PropertyKey
        {
            Fmtid = new Guid("1DA5D803-D492-4EDD-8C23-E0C0FFEE7F0E"),
            Pid = 0
        };

        public VolumeMixerManager()
        {
            try
            {
                _deviceEnumerator = (IMultiMediaDeviceEnumerator)new MMDeviceEnumerator();
            }
            catch (Exception ex)
            {
                throw new InvalidOperationException("Failed to initialize audio device enumerator. Make sure you're running on Windows Vista or later.", ex);
            }
        }

        public async Task<List<AudioDevice>> GetAudioDevicesAsync(DataFlow dataFlow = DataFlow.Output)
        {
            return await Task.Run(() =>
            {
                var devices = new List<AudioDevice>();

                if (_deviceEnumerator == null)
                {
                    return devices;
                }

                lock (_syncLock)
                {
                    try
                    {
                        var result = _deviceEnumerator.EnumAudioEndpoints(dataFlow, DeviceState.Active, out var deviceCollection);
                        if (result != 0 || deviceCollection == null)
                        {
                            return devices;
                        }

                        result = deviceCollection.GetCount(out var count);
                        if (result != 0)
                        {
                            ComHelper.ReleaseComObject(deviceCollection);
                            return devices;
                        }

                        string defaultId = "";
                        try
                        {
                            result = _deviceEnumerator.GetDefaultAudioEndpoint(dataFlow, Role.Multimedia, out var defaultDevice);
                            if (result == 0 && defaultDevice != null)
                            {
                                defaultDevice.GetId(out defaultId);
                                ComHelper.ReleaseComObject(defaultDevice);
                            }
                        }
                        catch
                        {
                            defaultId = "";
                        }

                        for (uint i = 0; i < count; i++)
                        {
                            try
                            {
                                result = deviceCollection.Item(i, out var device);
                                if (result == 0 && device != null)
                                {
                                    result = device.GetId(out var id);
                                    if (result == 0 && !string.IsNullOrEmpty(id))
                                    {
                                        var name = GetDeviceName(device);
                                        var type = GetDeviceType(device);
                                        bool isDefault = id == defaultId;
                                        devices.Add(new AudioDevice(device, id, name, isDefault, dataFlow, type));
                                    }
                                    else
                                    {
                                        ComHelper.ReleaseComObject(device);
                                    }
                                }
                            }
                            catch
                            {
                                // Do nothing
                            }
                        }

                        ComHelper.ReleaseComObject(deviceCollection);
                    }
                    catch
                    {
                        // Do nothing
                    }
                }

                return devices;
            });
        }

        public async Task<AudioDevice> GetDefaultAudioDeviceAsync(DataFlow dataFlow = DataFlow.Output)
        {
            return await Task.Run(() =>
            {
                if (_deviceEnumerator == null)
                {
                    return null;
                }

                lock (_syncLock)
                {
                    try
                    {
                        var result = _deviceEnumerator.GetDefaultAudioEndpoint(dataFlow, Role.Multimedia, out var device);
                        if (result == 0 && device != null)
                        {
                            result = device.GetId(out var id);
                            if (result == 0 && !string.IsNullOrEmpty(id))
                            {
                                var name = GetDeviceName(device);
                                var type = GetDeviceType(device);
                                return new AudioDevice(device, id, name, true, dataFlow, type);
                            }

                            ComHelper.ReleaseComObject(device);
                        }
                    }
                    catch
                    {
                        // Do nothing
                    }

                    return null;
                }
            });
        }

        private string GetDeviceName(IMultiMediaDevice device)
        {
            try
            {
                var result = device.OpenPropertyStore(0, out var propertyStore);
                if (result == 0 && propertyStore != null)
                {
                    var propertyKey = PKEY_Device_FriendlyName;
                    result = propertyStore.GetValue(ref propertyKey, out var propVariant);
                    if (result == 0 && propVariant.data1 != IntPtr.Zero)
                    {
                        string name = Marshal.PtrToStringUni(propVariant.data1);
                        ComHelper.ReleaseComObject(propertyStore);
                        return !string.IsNullOrEmpty(name) ? name : "Unknown Device";
                    }

                    ComHelper.ReleaseComObject(propertyStore);
                }
            }
            catch
            {
                // Do nothing
            }

            return "Unknown Device";
        }

        private DeviceType GetDeviceType(IMultiMediaDevice device)
        {
            try
            {
                int result = device.OpenPropertyStore(0, out var propertyStore);
                if (result == 0 && propertyStore != null)
                {
                    try
                    {
                        var propertyKey = PKEY_AudioEndpoint_FormFactor;
                        result = propertyStore.GetValue(ref propertyKey, out var propVariant);

                        // 0x13 == VT_UI4
                        if (result == 0 && propVariant.vt == 0x13)
                        {
                            int formFactor = propVariant.data1.ToInt32();

                            return formFactor switch
                            {
                                0 => DeviceType.Unknown,
                                1 => DeviceType.Speakers,
                                2 => DeviceType.LineLevel,
                                3 => DeviceType.Headphones,
                                4 => DeviceType.Microphone,
                                5 => DeviceType.Headset,
                                6 => DeviceType.Handset,
                                7 => DeviceType.UnknownDigitalPassthrough,
                                8 => DeviceType.SPDIF,
                                9 => DeviceType.DigitalAudioDisplayDevice,
                                10 => DeviceType.UnknownFormFactor,
                                11 => DeviceType.FMRadio,
                                12 => DeviceType.VideoPhone,
                                13 => DeviceType.RCA,
                                14 => DeviceType.Bluetooth,
                                15 => DeviceType.SPDIFOut,
                                16 => DeviceType.HDMI,
                                17 => DeviceType.DisplayAudio,
                                18 => DeviceType.UnknownFormFactor2,
                                19 => DeviceType.Other,
                                _ => DeviceType.Unknown,
                            };
                        }
                    }
                    finally
                    {
                        ComHelper.ReleaseComObject(propertyStore);
                    }
                }
            }
            catch
            {
                // Do nothing
            }

            return DeviceType.Unknown;
        }



        public async Task<bool> SetSystemVolumeAsync(float volume)
        {
            if (volume < 0f || volume > 1f)
            {
                return false;
            }

            AudioDevice defaultDevice = await GetDefaultAudioDeviceAsync(DataFlow.Output);

            if (defaultDevice == null)
            {
                return false;
            }

            try
            {
                return await defaultDevice.SetMasterVolumeAsync(volume);
            }
            finally
            {
                defaultDevice.Dispose();
            }
        }

        public async Task<float> GetSystemVolumeAsync()
        {
            AudioDevice defaultDevice = await GetDefaultAudioDeviceAsync(DataFlow.Output);

            if (defaultDevice == null)
            {
                return 0f;
            }

            try
            {
                return await defaultDevice.GetMasterVolumeAsync();
            }
            finally
            {
                defaultDevice.Dispose();
            }
        }

        public async Task<bool> SetSystemMuteAsync(bool mute)
        {
            AudioDevice defaultDevice = await GetDefaultAudioDeviceAsync(DataFlow.Output);

            if (defaultDevice == null)
            {
                return false;
            }

            try
            {
                return await defaultDevice.SetMasterMuteAsync(mute);
            }
            finally
            {
                defaultDevice.Dispose();
            }
        }

        public async Task<bool> GetSystemMuteAsync()
        {
            AudioDevice defaultDevice = await GetDefaultAudioDeviceAsync(DataFlow.Output);

            if (defaultDevice == null)
            {
                return false;
            }

            try
            {
                return await defaultDevice.GetMasterMuteAsync();
            }
            finally
            {
                defaultDevice.Dispose();
            }
        }

        public async Task<bool> SetMicrophoneVolumeAsync(float volume)
        {
            if (volume < 0f || volume > 1f)
            {
                return false;
            }

            AudioDevice defaultMic = await GetDefaultAudioDeviceAsync(DataFlow.Input);

            if (defaultMic == null)
            {
                return false;
            }

            try
            {
                return await defaultMic.SetMasterVolumeAsync(volume);
            }
            finally
            {
                defaultMic.Dispose();
            }
        }

        public async Task<float> GetMicrophoneVolumeAsync()
        {
            AudioDevice defaultMic = await GetDefaultAudioDeviceAsync(DataFlow.Input);

            if (defaultMic == null)
            {
                return 0f;
            }

            try
            {
                return await defaultMic.GetMasterVolumeAsync();
            }
            finally
            {
                defaultMic.Dispose();
            }
        }

        public async Task<bool> SetMicrophoneMuteAsync(bool mute)
        {
            AudioDevice defaultMic = await GetDefaultAudioDeviceAsync(DataFlow.Input);

            if (defaultMic == null)
            {
                return false;
            }

            try
            {
                return await defaultMic.SetMasterMuteAsync(mute);
            }
            finally
            {
                defaultMic.Dispose();
            }
        }

        public async Task<bool> GetMicrophoneMuteAsync()
        {
            AudioDevice defaultMic = await GetDefaultAudioDeviceAsync(DataFlow.Input);

            if (defaultMic == null)
            {
                return false;
            }

            try
            {
                return await defaultMic.GetMasterMuteAsync();
            }
            finally
            {
                defaultMic.Dispose();
            }
        }

        public async Task<bool> SetMicrophoneVolumeByNameAsync(string microphoneName, float volume)
        {
            if (string.IsNullOrWhiteSpace(microphoneName) || volume < 0f || volume > 1f)
            {
                return false;
            }

            List<AudioDevice> microphones = await GetAudioDevicesAsync(DataFlow.Input);

            foreach (var mic in microphones)
            {
                try
                {
                    if (mic.Name.IndexOf(microphoneName, StringComparison.OrdinalIgnoreCase) >= 0)
                    {
                        return await mic.SetMasterVolumeAsync(volume);
                    }
                }
                finally
                {
                    mic.Dispose();
                }
            }

            return false;
        }

        public async Task<List<AudioSession>> GetAllActiveSessionsAsync()
        {
            return await Task.Run(async () =>
            {
                var allSessions = new List<AudioSession>();

                List<AudioDevice> devices = await GetAudioDevicesAsync(DataFlow.Output);

                foreach (var device in devices)
                {
                    try
                    {
                        allSessions.AddRange(await device.GetAudioSessionsAsync());
                    }
                    catch
                    {
                        // Do nothing
                    }
                    finally
                    {
                        device.Dispose();
                    }
                }

                return allSessions;
            });
        }

        public async Task<List<AudioDevice>> GetMicrophonesAsync()
        {
            return await GetAudioDevicesAsync(DataFlow.Input);
        }

        public async Task<AudioDevice> GetDefaultMicrophoneAsync()
        {
            return await GetDefaultAudioDeviceAsync(DataFlow.Input);
        }

        public void Dispose()
        {
            lock (_syncLock)
            {
                if (!_isDisposed)
                {
                    ComHelper.ReleaseComObject(_deviceEnumerator);
                    _isDisposed = true;
                }
            }
        }
    }
}
