﻿using EonaCat.VolumeMixer.Helpers;
using EonaCat.VolumeMixer.Interfaces;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading.Tasks;

namespace EonaCat.VolumeMixer.Models
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.
    public class AudioDevice : IDisposable
    {
        internal Guid AudioController2Guid = new Guid("bfb7ff88-7239-4fc9-8fa2-07c950be9c6d");
        private readonly IMultiMediaDevice _device;
        private IAudioEndpointVolume _endpointVolume;
        private IAudioSessionManager _sessionManager;
        private bool _isDisposed = false;
        private readonly object _syncLock = new object();

        public string Id { get; private set; }
        public string Name { get; private set; }
        public DeviceType DeviceType { get; private set; }
        public bool IsDefault { get; private set; }
        public DataFlow DataFlow { get; private set; }

        internal AudioDevice(IMultiMediaDevice device, string id, string name, bool isDefault, DataFlow dataFlow, DeviceType deviceType)
        {
            _device = device;
            Id = id;
            Name = name;
            IsDefault = isDefault;
            DataFlow = dataFlow;
            DeviceType = deviceType;
            InitializeEndpointVolume();
            InitializeSessionManager();
        }

        private void InitializeEndpointVolume()
        {
            try
            {
                var audioEndpointGuid = typeof(IAudioEndpointVolume).GUID;
                var result = _device.Activate(ref audioEndpointGuid, 0, IntPtr.Zero, out var ptr);
                if (result == 0 && ptr != IntPtr.Zero)
                {
                    _endpointVolume = ComHelper.GetInterface<IAudioEndpointVolume>(ptr);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine($"Failed to initialize endpoint volume: {ex}");
                _endpointVolume = null;
            }
        }

        private void InitializeSessionManager()
        {
            try
            {
                var audioSessionGuid = typeof(IAudioSessionManager).GUID;
                var result = _device.Activate(ref audioSessionGuid, 0, IntPtr.Zero, out var ptr);
                if (result == 0 && ptr != IntPtr.Zero)
                {
                    _sessionManager = ComHelper.GetInterface<IAudioSessionManager>(ptr);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine($"Failed to initialize session manager: {ex}");
                _sessionManager = null;
            }
        }

        public async Task<float> GetMasterVolumeAsync()
        {
            return await Task.Run(() =>
            {
                lock (_syncLock)
                {
                    if (_isDisposed || _endpointVolume == null)
                    {
                        return 0f;
                    }

                    try
                    {
                        var result = _endpointVolume.GetMasterVolumeLevelScalar(out var volume);
                        return result == 0 ? volume : 0f;
                    }
                    catch (Exception ex)
                    {
                        Debug.WriteLine($"Error getting master volume: {ex}");
                        return 0f;
                    }
                }
            }).ConfigureAwait(false);
        }

        public async Task<bool> SetMasterVolumeAsync(float volume, int maxRetries = 5, int delayMs = 20)
        {
            if (_isDisposed || _endpointVolume == null || volume < 0f || volume > 1f)
            {
                return false;
            }

            var guid = Guid.Empty;

            for (int attempt = 0; attempt <= maxRetries; attempt++)
            {
                try
                {
                    int result;
                    lock (_syncLock)
                    {
                        result = _endpointVolume.SetMasterVolumeLevelScalar(volume, ref guid);
                    }

                    if (result == 0)
                    {
                        await Task.Delay(delayMs).ConfigureAwait(false);
                        var currentVolume = await GetMasterVolumeAsync().ConfigureAwait(false);
                        if (Math.Abs(currentVolume - volume) < 0.01f)
                        {
                            return true;
                        }
                    }
                }
                catch (Exception ex)
                {
                    Debug.WriteLine($"Volume set failed on attempt {attempt + 1}: {ex}");
                }

                await Task.Delay(delayMs).ConfigureAwait(false);
            }

            return false;
        }

        public async Task<bool> GetMasterMuteAsync()
        {
            return await Task.Run(() =>
            {
                lock (_syncLock)
                {
                    if (_isDisposed || _endpointVolume == null)
                    {
                        return false;
                    }

                    try
                    {
                        var result = _endpointVolume.GetMute(out var mute);
                        return result == 0 && mute;
                    }
                    catch (Exception ex)
                    {
                        Debug.WriteLine($"Error getting mute: {ex}");
                        return false;
                    }
                }
            }).ConfigureAwait(false);
        }

        public async Task<bool> SetMasterMuteAsync(bool mute)
        {
            if (_isDisposed || _endpointVolume == null)
            {
                return false;
            }

            try
            {
                var guid = Guid.Empty;
                return await Task.Run(() => _endpointVolume.SetMute(mute, ref guid) == 0).ConfigureAwait(false);
            }
            catch (Exception ex)
            {
                Debug.WriteLine($"Error setting mute: {ex}");
                return false;
            }
        }

        public async Task<List<AudioSession>> GetAudioSessionsAsync()
        {
            return await Task.Run(() =>
            {
                lock (_syncLock)
                {
                    var sessions = new List<AudioSession>();

                    if (_isDisposed || _sessionManager == null)
                    {
                        return sessions;
                    }

                    try
                    {
                        var result = _sessionManager.GetSessionEnumerator(out var sessionEnum);
                        if (result != 0 || sessionEnum == null)
                        {
                            return sessions;
                        }

                        result = sessionEnum.GetCount(out var count);
                        if (result != 0)
                        {
                            return sessions;
                        }

                        for (int i = 0; i < count; i++)
                        {
                            try
                            {
                                result = sessionEnum.GetSession(i, out var sessionControl);
                                if (result == 0 && sessionControl != null)
                                {
                                    var sessionControl2 = GetSessionControl2(sessionControl);
                                    if (sessionControl2 != null)
                                    {
                                        sessions.Add(new AudioSession(sessionControl2, _sessionManager));
                                        Marshal.ReleaseComObject(sessionControl2);
                                    }

                                    // Release the original sessionControl COM object
                                    Marshal.ReleaseComObject(sessionControl);
                                }
                            }
                            catch
                            {
                                // Do nothing
                            }
                        }
                    }
                    catch
                    {
                        // Do nothing
                    }

                    return sessions;
                }
            });
        }

        private IAudioSessionControlExtended GetSessionControl2(object sessionControl)
        {
            if (sessionControl == null)
            {
                return null;
            }

            var unknownPtr = Marshal.GetIUnknownForObject(sessionControl);
            try
            {
                IntPtr sessionControl2Ptr;
                int result = Marshal.QueryInterface(unknownPtr, ref AudioController2Guid, out sessionControl2Ptr);
                if (result == 0 && sessionControl2Ptr != IntPtr.Zero)
                {
                    return (IAudioSessionControlExtended)Marshal.GetObjectForIUnknown(sessionControl2Ptr);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine($"Error querying sessionControl2: {ex}");
            }
            finally
            {
                Marshal.Release(unknownPtr);
            }

            return null;
        }

        public async Task<bool> StepUpAsync(int delayMs = 20)
        {
            bool success = false;
            lock (_syncLock)
            {
                if (_isDisposed || _endpointVolume == null)
                {
                    return false;
                }

                try
                {
                    var guid = Guid.Empty;
                    success = _endpointVolume.StepUp(ref guid) == 0;
                }
                catch (Exception ex)
                {
                    Debug.WriteLine($"Error stepping up: {ex}");
                    success = false;
                }
            }

            await Task.Delay(delayMs).ConfigureAwait(false);
            return success;
        }

        public async Task<bool> StepDownAsync(int delayMs = 20)
        {
            bool success = false;
            lock (_syncLock)
            {
                if (_isDisposed || _endpointVolume == null)
                {
                    return false;
                }

                try
                {
                    var guid = Guid.Empty;
                    success = _endpointVolume.StepDown(ref guid) == 0;
                }
                catch (Exception ex)
                {
                    Debug.WriteLine($"Error stepping down: {ex}");
                    success = false;
                }
            }

            await Task.Delay(delayMs).ConfigureAwait(false);
            return success;
        }

        public void Dispose()
        {
            lock (_syncLock)
            {
                if (_isDisposed)
                {
                    return;
                }

                ComHelper.ReleaseComObject(_endpointVolume);
                ComHelper.ReleaseComObject(_sessionManager);
                ComHelper.ReleaseComObject(_device);
                _isDisposed = true;

                GC.SuppressFinalize(this);
            }
        }
    }
}
