﻿using EonaCat.VolumeMixer.Helpers;
using EonaCat.VolumeMixer.Interfaces;
using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading.Tasks;

namespace EonaCat.VolumeMixer.Models
{
    public class AudioSession : IDisposable
    {
        private readonly IAudioSessionControlExtended _sessionControl;
        private IAudioVolume _audioVolume;
        private bool _isDisposed = false;
        private readonly object _syncLock = new();

        public string DisplayName { get; private set; }
        public string IconPath { get; private set; }
        public uint ProcessId { get; private set; }
        public AudioSessionState State { get; private set; }

        internal AudioSession(IAudioSessionControlExtended sessionControl, IAudioSessionManager sessionManager)
        {
            _sessionControl = sessionControl ?? throw new ArgumentNullException(nameof(sessionControl));
            InitializeSimpleAudioVolume();
            LoadSessionInfo();
        }

        private void InitializeSimpleAudioVolume()
        {
            lock (_syncLock)
            {
                try
                {
                    var sessionControlUnknown = Marshal.GetIUnknownForObject(_sessionControl);
                    var iidSimpleAudioVolume = typeof(IAudioVolume).GUID;
                    IntPtr simpleAudioVolumePtr = IntPtr.Zero;

                    int result = Marshal.QueryInterface(sessionControlUnknown, ref iidSimpleAudioVolume, out simpleAudioVolumePtr);
                    if (result == 0 && simpleAudioVolumePtr != IntPtr.Zero)
                    {
                        _audioVolume = (IAudioVolume)Marshal.GetObjectForIUnknown(simpleAudioVolumePtr);
                        Marshal.Release(simpleAudioVolumePtr);
                    }
                    Marshal.Release(sessionControlUnknown);
                }
                catch
                {
                    _audioVolume = null;
                }
            }
        }

        private void LoadSessionInfo()
        {
            lock (_syncLock)
            {
                try
                {
                    var result = _sessionControl.GetDisplayName(out var displayName);
                    DisplayName = (result == 0 && !string.IsNullOrEmpty(displayName)) ? displayName : string.Empty;

                    result = _sessionControl.GetIconPath(out var iconPath);
                    IconPath = (result == 0) ? (iconPath ?? string.Empty) : string.Empty;

                    result = _sessionControl.GetProcessId(out var processId);
                    ProcessId = (result == 0) ? processId : 0;

                    result = _sessionControl.GetState(out var state);
                    State = (result == 0) ? state : AudioSessionState.AudioSessionStateInactive;

                    if (string.IsNullOrEmpty(DisplayName) && ProcessId != 0)
                    {
                        try
                        {
                            var process = Process.GetProcessById((int)ProcessId);
                            DisplayName = process.ProcessName;
                        }
                        catch
                        {
                            DisplayName = "Unknown";
                        }
                    }

                    if (string.IsNullOrEmpty(IconPath) && ProcessId != 0)
                    {
                        try
                        {
                            var process = Process.GetProcessById((int)ProcessId);
                            IconPath = process.MainModule?.FileName ?? string.Empty;
                        }
                        catch
                        {
                            IconPath = "Unknown";
                        }
                    }

                    if (ProcessId == 0 && _sessionControl.IsSystemSoundsSession() == 0)
                    {
                        DisplayName = "System sounds";
                        IconPath = "Unknown";
                    }
                }
                catch
                {
                    DisplayName = string.Empty;
                    IconPath = string.Empty;
                    ProcessId = 0;
                    State = AudioSessionState.AudioSessionStateInactive;
                }
            }
        }

        public async Task<float> GetVolumeAsync()
        {
            lock (_syncLock)
            {
                if (_isDisposed || _audioVolume == null)
                {
                    return 0f;
                }
                try
                {
                    int result = _audioVolume.GetMasterVolume(out var volume);
                    return result == 0 ? volume : 0f;
                }
                catch
                {
                    return 0f;
                }
            }
        }

        public async Task<bool> SetVolumeAsync(float volume, int maxRetries = 2, int delayMs = 20)
        {
            if (volume < 0f || volume > 1f)
            {
                return false;
            }

            IAudioVolume simpleAudioVolCopy;

            lock (_syncLock)
            {
                if (_isDisposed || _audioVolume == null)
                {
                    return false;
                }

                simpleAudioVolCopy = _audioVolume;
            }

            var guid = Guid.Empty;

            for (int attempt = 0; attempt <= maxRetries; attempt++)
            {
                try
                {
                    var result = simpleAudioVolCopy.SetMasterVolume(volume, ref guid);
                    if (result == 0)
                    {
                        await Task.Delay(delayMs).ConfigureAwait(false);

                        var currentVolume = await GetVolumeAsync().ConfigureAwait(false);
                        if (Math.Abs(currentVolume - volume) < 0.01f)
                        {
                            return true;
                        }
                    }
                }
                catch
                {
                    // Retry
                }
                await Task.Delay(delayMs).ConfigureAwait(false);
            }

            return false;
        }

        public async Task<bool> GetMuteAsync()
        {
            lock (_syncLock)
            {
                if (_isDisposed || _audioVolume == null)
                {
                    return false;
                }

                try
                {
                    var result = _audioVolume.GetMute(out var mute);
                    return result == 0 && mute;
                }
                catch
                {
                    return false;
                }
            }
        }

        public async Task<bool> SetMuteAsync(bool mute)
        {
            IAudioVolume simpleAudioVolCopy;

            lock (_syncLock)
            {
                if (_isDisposed || _audioVolume == null)
                {
                    return false;
                }

                simpleAudioVolCopy = _audioVolume;
            }

            try
            {
                var guid = Guid.Empty;
                return await Task.Run(() => simpleAudioVolCopy.SetMute(mute, ref guid) == 0).ConfigureAwait(false);
            }
            catch
            {
                return false;
            }
        }

        public async Task<string> GetProcessNameAsync()
        {
            if (ProcessId == 0)
            {
                return "Unknown";
            }

            try
            {
                var process = Process.GetProcessById((int)ProcessId);
                return process.ProcessName;
            }
            catch
            {
                return "Unknown";
            }
        }

        public async Task<float> GetEffectiveVolumeAsync(AudioDevice device)
        {
            var deviceVolume = await device.GetMasterVolumeAsync().ConfigureAwait(false);
            var sessionVolume = await GetVolumeAsync().ConfigureAwait(false);
            return deviceVolume * sessionVolume;
        }

        public void Dispose()
        {
            lock (_syncLock)
            {
                if (_isDisposed)
                {
                    return;
                }

                if (_audioVolume != null)
                {
                    Marshal.ReleaseComObject(_audioVolume);
                    _audioVolume = null;
                }
                ComHelper.ReleaseComObject(_sessionControl);

                _isDisposed = true;
            }
        }
    }
}
